--
-- This is a win32 optimised version of directory listing
--

with
   System,
   Win32.Winnt,
   Win32.Winbase,
   Ada.Exceptions,
   Win32,
   System.Bit_Ops,
   File_Utilities,
   Paskma.Logger,
   C_Conversions,
   Ada.Unchecked_Conversion;

use System,
   Win32,
   Win32.Winbase,
   Ada.Exceptions,
   System.Bit_Ops,
   File_Utilities,
   Paskma.Logger,
   C_Conversions,
   Win32.Winnt;

package body Directory_Listings is

   use Directory_Item_Lists;

   --
   -- To Directory Item
   --

   function To_Directory_Item (
         Found : Win32_Find_Data ) 
     return Directory_Item_Type is 

      Result          : Directory_Item_Type;  
      Tmp_Size_Low,  
      Tmp_Size_High   : Long_Long_Integer;  
      Bitresult,  
      Tmp_Attribute1,  
      Tmp_Attribute2  : aliased Win32.Dword;  
      Tmp_Source_Time : aliased Filetime;  
      Tmp_Result_Time : aliased Systemtime;  
      Suc             : Win32.Bool;  
   begin
      --
      -- File Name
      Result.Name := To_Unbounded_String(To_String(Found.Cfilename));
      --
      -- File Size
      Tmp_Size_Low := Long_Long_Integer(Found.Nfilesizelow);
      Tmp_Size_High := Long_Long_Integer(Found.Nfilesizehigh);
      Result.Size := Tmp_Size_High * 2**32 + Tmp_Size_Low;
      --
      -- Attributes
      for I in Result.Attributes'range loop
         Result.Attributes(I) := False;
      end loop;
      --
      -- directory
      Tmp_Attribute1 := File_Attribute_Directory;
      Tmp_Attribute2 :=  Found.Dwfileattributes;

      Bit_And(Tmp_Attribute1'Address, Tmp_Attribute1'Size,
         Tmp_Attribute2'Address, Tmp_Attribute2'Size,
         Bitresult'Address);

      if Integer(Bitresult) /= 0 then
         Result.Attributes(Directory) := True;
      end if;
      --
      -- readonly
      Tmp_Attribute1 := File_Attribute_Readonly;
      Tmp_Attribute2 :=  Found.Dwfileattributes;

      Bit_And(Tmp_Attribute1'Address, Tmp_Attribute1'Size,
         Tmp_Attribute2'Address, Tmp_Attribute2'Size,
         Bitresult'Address);

      if Integer(Bitresult) /= 0 then
         Result.Attributes(Read_Only) := True;
      end if;

      --
      -- Creation time
      Tmp_Source_Time := Found.Ftlastwritetime;
      Suc := Filetimetosystemtime(Tmp_Source_Time'Unchecked_Access,
         Tmp_Result_Time'Unchecked_Access);

      if Suc = 0 then
         raise Listing_Error;
      end if;

      Result.Creation_Time := Time_Of(
         Year    => Year_Number (Tmp_Result_Time.Wyear),                                                                                               
         Month   => Month_Number (Tmp_Result_Time.Wmonth),                                                                                             
         Day     => Day_Number (Tmp_Result_Time.Wday),                                                                                                 
         Seconds => Day_Duration (Integer (Tmp_Result_Time.Whour) * 3600 + Integer (Tmp_Result_Time.Wminute) * 60 + Integer (Tmp_Result_Time.Wsecond))
         );

      return Result;
   end;


   -----------------
   -- Get_Listing --
   -----------------

   procedure Get_Listing (
         Directory_Name  :        Unbounded_String;                 
         Result          : in out Directory_Item_Lists.List;        
         Attach_Asterisk :        Boolean                   := True ) is 

      Listing_Handle          : Handle;  
      Internal_Directory_Name : Unbounded_String;  
      Found                   : aliased Win32_Find_Data;  
      Usuable_Found           : Directory_Item_Type;  
      Suc                     : Bool;  
   begin
      Empty(Result);

      if Attach_Asterisk then
         Internal_Directory_Name := Compose_Path(Directory_Name,
            To_Unbounded_String("*"&Character'First));
      else
         -- this is usuable when we want determine attributes of given filename only
         Internal_Directory_Name := Directory_Name & Character'First;
         --attach '\0'
      end if;

      --Log("Get_Listing for " & Internal_Directory_Name);

      Listing_Handle := Findfirstfile(To_Lpcstr(To_String(
               Internal_Directory_Name)),
         Found'Unchecked_Access);

      if Listing_Handle = Invalid_Handle_Value then
         -- probaly not valid filename was given
         -- raise Listing_Error;
         Raise_Exception(Listing_Error'Identity, "Probably invalid filename " & To_String(Directory_Name));
      end if;

      loop
         Usuable_Found := To_Directory_Item(Found);
         if Usuable_Found.Name /= "." and Usuable_Found.Name /= ".." then
            -- add to the list only if it is the interesting filename
            Push_Back(Usuable_Found, Result);
         end if;
         Suc := Findnextfile(Listing_Handle, Found'Unchecked_Access);
         exit when Suc = 0;
      end loop;

      Suc := Findclose(Listing_Handle);
      if Suc = 0 then
         raise Listing_Error;
      end if;

   end Get_Listing;

   --
   -- Is_Directory
   --

   function Is_Directory (
         Name : String ) 
     return Boolean is 

      --Invalid_File_Attributes : constant Win32.Dword         := Win32.Dword (16#ffffffff#);  
      Result_Attribute,  
      Mask_Attribute,  
      Bitresult        : aliased Win32.Dword;  
      Internal_Name    : String (Name'First .. Name'Last + 1) := Name & Character'First;  

      function Convert is 
      new Ada.Unchecked_Conversion(Win32.Dword, Integer);
   begin

      Result_Attribute := Getfileattributes(To_Lpcstr(Internal_Name));
      Mask_Attribute := File_Attribute_Directory;

      if Convert(Result_Attribute) = -1 then
         return False;
      end if;

      Bit_And(Result_Attribute'Address, Result_Attribute'Size,
         Mask_Attribute'Address, Mask_Attribute'Size,
         Bitresult'Address);

      if Integer(Bitresult) = 0 then
         return False;
      else
         return True;
      end if;

   end;

end Directory_Listings;

