with
   Pi_Conversion_Layer,
   Reply_Codes,
   Pi_Commands,
   Port_Generator,
   Socket_Utilities,
   Ada.Streams.Stream_Io,
   Ada.Exceptions,
   Paskma.Time_Utilities,
   Ada.Strings.Unbounded;

use
   Pi_Conversion_Layer,
   Reply_Codes,
   Pi_Commands,
   Port_Generator,
   Socket_Utilities,
   Ada.Streams.Stream_Io,
   Ada.Exceptions,
   Paskma.Time_Utilities,
   Ada.Strings.Unbounded;

package body Ftp_Protocol_Interpreter is

   use Unbounded_String_Lists;

   procedure Reset (
         Obj : in out Pi_Context_Type ); 

   -------------------------------------------------------------------------------------

   procedure Initialize (
         Obj : in out Pi_Context_Type ) is 
   begin
      Reset(Obj);
      Obj.Connection_Mode := Active;
      Obj.Socket := No_Socket;
      Obj.Task_Id := -1;
   end Initialize;

   -------------------------------------------------------------------------------------

   procedure Construct (
         Obj                    : in out Pi_Context_Type;          
         Ftp_Configuration      :        Ftp_Configuration_Type;   
         Socket                 :        Socket_Type;              
         Task_Id                :        Task_Id_Type;             
         A_Connected_To_Address :        Sock_Addr_Type;           
         A_Statistics           :        Statistics_Manager_Access ) is 
   begin
      Obj.Socket := Socket;
      Obj.Configuration := Ftp_Configuration;
      Obj.Task_Id := Task_Id;
      Obj.Connected_To_Address := A_Connected_To_Address;
      Obj.Statistics := A_Statistics;
      -- other fields were initialized by Initialize procedure
   end Construct;
   -------------------------------------------------------------------------------------

   --
   -- Send_Reply
   --

   procedure Send_Reply (
         Obj         : Pi_Context_Type; 
         Of_Code     : Reply_Code_Type; 
         And_Message : String           ) is 
   begin
      Send_Reply(Obj.Socket, Of_Code, And_Message);
      Log(Obj.Task_Id, Of_Code, And_Message);
   end Send_Reply;

   procedure Send_Reply (
         Obj         : Pi_Context_Type; 
         Of_Code     : Reply_Code_Type; 
         And_Message : Unbounded_String ) is 
   begin
      Send_Reply(Obj.Socket, Of_Code, And_Message);
      Log(Obj.Task_Id, Of_Code, And_Message);
   end Send_Reply;

   --
   -- Reset
   --
   -- Resets the context to the initial state (no user logged, no file_tree assigned, active connection mode...
   -- But socket is still open!

   procedure Reset (
         Obj : in out Pi_Context_Type ) is 
   begin
      Uninitialize(Obj.File_Tree);
      Obj.Log_State := Unlogged;
      Obj.Command_State := Uninteresting;
      Obj.Username := To_Unbounded_String("");
      Obj.Password := To_Unbounded_String("");
   end Reset;

   -------------------------------------------------------------------------------------

   procedure Write_Welcome_Message (
         Obj : Pi_Context_Type ) is 

      Message : constant String := "Paskma FTP Server Ready";  
   begin
      Send_Reply(Obj, The_220, Message);
   end Write_Welcome_Message;

   -------------------------------------------------------------------------------------

   procedure Send_Not_Logged_In (
         Obj : Pi_Context_Type ) is 

      Message_Not_Logged : constant String := "Not logged in.";  
   begin
      Send_Reply(Obj, The_550, Message_Not_Logged);
   end Send_Not_Logged_In;

   -------------------------------------------------------------------------------------

   procedure Command_User (
         Obj       : in out Pi_Context_Type; 
         Parameter :        Unbounded_String ) is 

      Message_Ok        : constant String := "Username OK, need password.";  
      Message_Invalid_1 : constant String := "Invalid username (";  
      Message_Invalid_2 : constant String := "), not logged in.";  
   begin
      if Obj.Log_State /= Unlogged then
         -- if the state is not initial then reset to initial
         Obj.Statistics.Log_Out_User(Obj.Task_Id);
         Reset(Obj);
      end if;
      --
      -- USER command can occur in every state
      Verify_User(Obj.Configuration, Parameter);
      -- exception when username is not in the database
      Obj.Username := Parameter;
      Obj.Command_State := Waiting_Pass;
      -- note: log_state unchanged until valid password is given
      Send_Reply(Obj, The_331, Message_Ok);
   exception
      when Ftp_Configurations.Access_Not_Allowed =>
         Send_Reply(Obj, The_530, Message_Invalid_1 & Parameter & Message_Invalid_2);
   end Command_User;

   -------------------------------------------------------------------------------------

   procedure Command_Pass (
         Obj      : in out Pi_Context_Type; 
         Password :        Unbounded_String ) is 

      Message_Invalid_Sequence : constant String := "Invalid sequence of commands. USER must be before PASS.";  
      Message_Logged_In        : constant String := "User logged in. Welcome! Proceed.";  
      Message_Invalid_Password : constant String := "Invalid password. Not logged in.";  
   begin
      if Obj.Command_State /= Waiting_Pass then
         -- state is not valid for accepting password
         Reset(Obj);
         Send_Reply(Obj, The_503, Message_Invalid_Sequence);
      else
         begin
            -- state valid
            Verify_User(Obj.Configuration, Obj.Username, Password);
            -- Is the password ok? If not then exception is raised            

            --
            -- user logged, create his file tree
            Obj.Password := Password;
            Obj.Log_State := Logged;

            Obj.Statistics.Log_In_User(
               Obj.Username,
               Obj.Task_Id,
               Obj.Connected_To_Address);

            Obj.Command_State := Uninteresting;
            Obj.File_Tree := Get_File_Tree(
               Obj.Configuration,
               Obj.Username,
               Obj.Password);

            Send_Reply(Obj, The_230, Message_Logged_In);
         exception
            when Ftp_Configurations.Access_Not_Allowed =>
               Send_Reply(Obj, The_530, Message_Invalid_Password);
         end;
      end if;
   end Command_Pass;

   -------------------------------------------------------------------------------------

   procedure Command_Quit (
         Obj : in out Pi_Context_Type ) is 
      Message_Closing : constant String := "Server is closing the connection. Have a nice day!";  
   begin
      begin
         Obj.Statistics.Log_Out_User(Obj.Task_Id);
      exception
         when Event : Item_Not_Found
               =>
            Log(Obj.Task_Id, "Premature user logout - statistic item not found");
      end;
      Send_Reply(Obj, The_221, Message_Closing);
      Close_Socket(Obj.Socket);
      Reset(Obj); -- not needed indeed, this task will be ended, but...
   end Command_Quit;

   -------------------------------------------------------------------------------------

   procedure Command_Pwd (
         Obj : in out Pi_Context_Type ) is 

      Working_Directory,  
      Message_Success     :          Unbounded_String;  
      Message_Success_End : constant String           := " is current directory.";  
   begin
      if Obj.Log_State = Logged then
         Working_Directory := Get_Current_Directory(Obj.File_Tree);

         Message_Success := '"' & Working_Directory &  '"' &
            Message_Success_End;

         Send_Reply(Obj, The_257, Message_Success);
      else
         Send_Not_Logged_In(Obj);
      end if;

   end Command_Pwd;

   -------------------------------------------------------------------------------------

   procedure Command_Cdup (
         Obj : in out Pi_Context_Type ) is 

      Message_Ok       :          Unbounded_String;  
      Message_Ok_Begin : constant String           := "Directory changed to: ";  
   begin
      if Obj.Log_State = Logged then

         Change_Directory_Up(Obj.File_Tree);

         Message_Ok := Message_Ok_Begin & '"' & Get_Current_Directory(Obj.
            File_Tree) & '"';
         Send_Reply(Obj, The_250,Message_Ok);
      else
         Send_Not_Logged_In(Obj);
      end if;
   end Command_Cdup;

   -------------------------------------------------------------------------------------

   procedure Command_Cwd (
         Obj            : in out Pi_Context_Type; 
         Directory_Name :        Unbounded_String ) is 

      Message_Ok,  
      Message_Invalid_Filename     :          Unbounded_String;  
      Message_Ok_Begin             : constant String           := "Directory changed to: ";  
      Message_Invalid_Filename_End : constant String           := " is not valid directory name.";  
   begin
      if Obj.Log_State = Logged then

         -- TODO: this cases can solve lower layer (ftp tree)
         if Directory_Name = "." then
            null; -- only write the OK message below
         elsif  Directory_Name = ".." then
            -- the "CDUP" command is in RFC because not every OS has ".." directory item
            -- But some FTP clients (like the included in windows) are stupid.
            Command_Cdup(Obj);
            return;
         else
            Change_Directory(Obj.File_Tree, Directory_Name);
         end if;

         Message_Ok := Message_Ok_Begin & '"' & Get_Current_Directory(Obj.File_Tree) & '"';
         Send_Reply(Obj, The_250, Message_Ok);
      else
         Send_Not_Logged_In(Obj);
      end if;
   exception
      when Invalid_Filename =>
         Message_Invalid_Filename := '"' & Directory_Name & '"' & Message_Invalid_Filename_End;
         Send_Reply(Obj, The_550, Message_Invalid_Filename);
   end Command_Cwd;

   -------------------------------------------------------------------------------------

   procedure Command_Port (
         Obj              : in out Pi_Context_Type; 
         Address_And_Port :        Unbounded_String ) is 

      Message_Ok             : constant String := "PORT Command successful.";  
      Message_Not_Recognized : constant String := "Address not recognized.";  
   begin
      if Obj.Log_State = Logged then
         Obj.Connection_Mode := Active; -- command port means active mode
         Obj.Remote_Address := Ftp_Address_To_Socket_Address(
            Address_And_Port);

         Send_Reply(Obj, The_200, Message_Ok);
         Log(Obj.Task_Id, "Remote address changed to: " & Image(Obj.
               Remote_Address));
      else
         Send_Not_Logged_In(Obj);
      end if;

   exception
      when Constraint_Error =>
         Send_Reply(Obj, The_501, Message_Not_Recognized);

   end Command_Port;


   -------------------------------------------------------------------------------------

   procedure Command_Pasv (
         Obj : in out Pi_Context_Type ) is 
      Message_Abort : constant String := "Error creating listen socket, passive mode not set";  
   begin
      if Obj.Log_State = Logged then
         begin
            Obj.Connection_Mode := Passive;

            Obj.Passive_Listen_Address.Addr := Get_Ip(Obj.Configuration);
            The_Port_Generator.Get_New_Port(Obj.Passive_Listen_Address.Port);

            --
            -- This code

            --            begin
            --               if Obj.Passive_Listen_Socket /= No_Socket then
            --                  Log(Obj.Task_Id, "Closing old passv socket.");
            --                  -- this situation may appear when client sends multiple PASV commands.
            --                  -- so this brach closes the unclosed sockets
            --                  Close_Socket(Obj.Passive_Listen_Socket);
            --               end if;
            --            exception
            --               when Event : Socket_Error =>
            --                  Log(Obj.Task_Id, "Exception raised - "  & Exception_Information(Event));
            --                  Log(Obj.Task_Id, "Socket image is " & Image(Obj.Passive_Listen_Socket));
            --            end;

            Create_Socket(Obj.Passive_Listen_Socket);

            Bind_Socket(Obj.Passive_Listen_Socket, Obj.Passive_Listen_Address);
            Listen_Socket(Obj.Passive_Listen_Socket);

            Log(Obj.Task_Id, "Listening on address: " & Image(Obj.Passive_Listen_Address));

            Send_Reply(Obj, The_227,  "Entering Passive Mode (" & Socket_Address_To_Ftp_Address(Obj.Passive_Listen_Address) & ")");
         exception
            when Event : Socket_Error=>
               -- error listening on the socket
               Send_Reply(Obj, The_502, Message_Abort);
               Log(Obj.Task_Id, "Error creating passive listen socket - " & Exception_Information(Event));
               Obj.Connection_Mode := Active;
         end;
      else
         Send_Not_Logged_In(Obj);
      end if;
   end;

   -------------------------------------------------------------------------------------

   procedure Command_List (
         Obj            : in out Pi_Context_Type;          
         Directory_Name :        Unbounded_String;         
         Names_Only     :        Boolean          := False ) is 
      -- if names_only is TRUE then this command is NLST 

      The_Listing : Unbounded_String_Lists.List;  
      I,  
      I_End       : Unbounded_String_Lists.Iterator;  
      Data_Socket : Socket_Type;  

      Message_Openning    : constant String := "Command OK. Openning data connection for response.";  
      Message_Closed      : constant String := "Transfer OK. Data connection closed.";  
      Message_Cannot_Open : constant String := "Cannot open data connection.";  
      Message_Abort       : constant String := "Transfer aborted.";  
   begin
      if Obj.Log_State = Logged then

         Obj.Statistics.Set_Activity(Obj.Task_Id, "Listing - " &
            Get_Absolut_Name(Obj.File_Tree, Directory_Name));

         Get_Directory_Listing(Obj.File_Tree, Directory_Name, The_Listing,
            Names_Only);

         Send_Reply(Obj, The_150, Message_Openning);
         begin
            if Obj.Connection_Mode = Active then
               Create_Socket(Data_Socket);
               Connect_Socket(Data_Socket, Obj.Remote_Address);
            else
               Accept_Socket(Obj.Passive_Listen_Socket, Data_Socket, Obj.Remote_Address);
               Close_Socket(Obj.Passive_Listen_Socket);
               Log(Obj.Task_Id, "Remote address changed to, listen socket closed: " & Image(Obj.Remote_Address));
            end if;
         exception
            -- the given port is probably already used
            when Socket_Error =>
               Send_Reply(Obj, The_425, Message_Cannot_Open);
               return;
         end;

         I:= Get_Begin(The_Listing);
         I_End := Get_End(The_Listing);
         while I /= I_End loop
            Write_Line(Data_Socket, Dereference(I));
            Increment(I);
         end loop;

         Close_Socket(Data_Socket);

         Send_Reply(Obj, The_226, Message_Closed);
      else
         Send_Not_Logged_In(Obj);
      end if;
   exception
      when Socket_Error=>
         -- the socket was closed by client
         Send_Reply(Obj, The_426, Message_Abort);
   end Command_List;

   -------------------------------------------------------------------------------------

   procedure Command_Retr (
         Obj      : in out Pi_Context_Type; 
         Filename :        Unbounded_String ) is 

      Message_Openning         : constant String := "Command OK. Openning data connection for response.";  
      Message_Closed           : constant String := "Transfer OK. Data connection closed.";  
      Message_Cannot_Open      : constant String := "Cannot open data connection.";  
      Message_Abort            : constant String := "Transfer aborted.";  
      Message_Not_Allowed      : constant String := "Access not allowed.";  
      Message_Invalid_Filename : constant String := "Invalid filename: ";  
      Message_End_Error        : constant String := "Unexpected file end. Probaly wrong restart.";  

      File        : File_Type;  
      Data_Socket : Socket_Type;  
   begin
      if Obj.Log_State = Logged then

         Obj.Statistics.Set_Activity(Obj.Task_Id, "Retrieving - " &
            Get_Absolut_Name(Obj.File_Tree, Filename));

         Open_File_Read(Obj.File_Tree, File, Filename);

         Send_Reply(Obj, The_150, Message_Openning);

         begin
            if Obj.Connection_Mode = Active then
               Create_Socket(Data_Socket);
               Connect_Socket(Data_Socket, Obj.Remote_Address);
            else
               Accept_Socket(Obj.Passive_Listen_Socket, Data_Socket, Obj.Remote_Address);
               Close_Socket(Obj.Passive_Listen_Socket);
               Log(Obj.Task_Id, "Remote address changed to, listen socket closed: " & Image(Obj.Remote_Address));
            end if;
         exception
            when Socket_Error =>
               --
               -- the given port is probably already used
               -- tidy up and exit
               Send_Reply(Obj, The_425, Message_Cannot_Open);
               Close(File);
               return;
         end;

         --
         -- Restart
         if Obj.Command_State = Restart_Marker_Set then
            -- client wants to restart aborted download
            Set_Index(File, Obj.Restart_Marker_Index);
            -- if the index is too far, the end_error will be raised by read procedure
            Obj.Command_State := Uninteresting;
         end if;

         Copy_File_To_Socket(File, Data_Socket);
         Close_Socket(Data_Socket);
         Close(File);

         Send_Reply(Obj, The_226, Message_Closed);
      else
         Send_Not_Logged_In(Obj);
      end if;

   exception
      when File_Trees.Access_Not_Allowed =>
         Send_Reply(Obj, The_550, Message_Not_Allowed);
      when Invalid_Filename =>
         Send_Reply(Obj, The_550, Message_Invalid_Filename & Filename);
      when Event : Socket_Error=>
         -- the socket was closed by client
         Send_Reply(Obj, The_426, Message_Abort);
         Log(Obj.Task_Id, "Socket closed by peer - " & Exception_Information(Event));
         Close(File);
      when End_Error =>
         Send_Reply(Obj, The_450, Message_End_Error);
         Close(File);
         Close_Socket(Data_Socket);
   end Command_Retr;

   -------------------------------------------------------------------------------------

   procedure Command_Stor (
         Obj      : in out Pi_Context_Type; 
         Filename :        Unbounded_String ) is 

      Message_Openning         : constant String := "Command OK. Openning data connection for response.";  
      Message_Closed           : constant String := "Transfer OK. Data connection closed.";  
      Message_Cannot_Open      : constant String := "Cannot open data connection.";  
      Message_Abort            : constant String := "Transfer aborted.";  
      Message_Not_Allowed      : constant String := "Access not allowed.";  
      Message_Invalid_Filename : constant String := "Invalid filename: ";  

      File        : File_Type;  
      Data_Socket : Socket_Type;  
   begin
      if Obj.Log_State = Logged then

         Obj.Statistics.Set_Activity(Obj.Task_Id, "Storing - " &
            Get_Absolut_Name(Obj.File_Tree, Filename));

         --
         -- Opening the file
         if Obj.Command_State = Restart_Marker_Set then
            -- Restart support
            Obj.Command_State := Uninteresting;
            Open_File_Read_Write(Obj.File_Tree, File, Filename);
            -- also delete permition needed
            Set_Index(File, Obj.Restart_Marker_Index);
            Set_Mode(File, Out_File);
         else
            -- Ordinary upload
            Open_File_Write(Obj.File_Tree, File, Filename);
         end if;

         Send_Reply(Obj, The_150, Message_Openning);

         begin
            if Obj.Connection_Mode = Active then
               Create_Socket(Data_Socket);
               Connect_Socket(Data_Socket, Obj.Remote_Address);
            else
               Accept_Socket(Obj.Passive_Listen_Socket, Data_Socket, Obj.Remote_Address);
               Close_Socket(Obj.Passive_Listen_Socket);
               Log(Obj.Task_Id, "Remote address changed to, listen socket closed: " & Image(Obj.Remote_Address));
            end if;

         exception
            when Socket_Error =>
               --
               -- the given port is probably already used
               -- tidy up and exit
               Send_Reply(Obj, The_425, Message_Cannot_Open);
               Close(File);
               return;
         end;

         Copy_Socket_To_File(Data_Socket, File);
         Close_Socket(Data_Socket);
         Close(File);

         Send_Reply(Obj, The_226, Message_Closed);
      else
         Send_Not_Logged_In(Obj);
      end if;

   exception
      when File_Trees.Access_Not_Allowed =>
         Send_Reply(Obj, The_550, Message_Not_Allowed);
      when Invalid_Filename =>
         Send_Reply(Obj, The_550, Message_Invalid_Filename & Filename);
      when Event : Socket_Error=>
         -- the socket was closed by client
         Send_Reply(Obj, The_426, Message_Abort);
         Log(Obj.Task_Id, Exception_Information(Event));
         Close(File);
   end Command_Stor;

   -------------------------------------------------------------------------------------

   procedure Command_Appe (
         Obj      : in out Pi_Context_Type; 
         Filename :        Unbounded_String ) is 

      Message_Openning         : constant String := "Command OK. Openning data connection for response.";  
      Message_Closed           : constant String := "Transfer OK. Data connection closed.";  
      Message_Cannot_Open      : constant String := "Cannot open data connection.";  
      Message_Abort            : constant String := "Transfer aborted.";  
      Message_Not_Allowed      : constant String := "Access not allowed.";  
      Message_Invalid_Filename : constant String := "Invalid filename: ";  

      File        : File_Type;  
      Data_Socket : Socket_Type;  
   begin
      if Obj.Log_State = Logged then

         Obj.Statistics.Set_Activity(Obj.Task_Id, "Appending - " &
            Get_Absolut_Name(Obj.File_Tree, Filename));

         --
         -- Opening the file
         Open_File_Append(Obj.File_Tree, File, Filename);
         -- if the file does not exist it will be created

         Send_Reply(Obj, The_150, Message_Openning);

         begin
            if Obj.Connection_Mode = Active then
               Create_Socket(Data_Socket);
               Connect_Socket(Data_Socket, Obj.Remote_Address);
            else
               Accept_Socket(Obj.Passive_Listen_Socket, Data_Socket, Obj.Remote_Address);
               Close_Socket(Obj.Passive_Listen_Socket);
               Log(Obj.Task_Id, "Remote address changed to, listen socket closed: " & Image(Obj.Remote_Address));
            end if;

         exception
            when Socket_Error =>
               --
               -- the given port is probably already used
               -- tidy up and exit
               Send_Reply(Obj, The_425, Message_Cannot_Open);
               Close(File);
               return;
         end;

         Copy_Socket_To_File(Data_Socket, File);
         Close_Socket(Data_Socket);
         Close(File);

         Send_Reply(Obj, The_226, Message_Closed);
      else
         Send_Not_Logged_In(Obj);
      end if;

   exception
      when File_Trees.Access_Not_Allowed =>
         Send_Reply(Obj, The_550, Message_Not_Allowed);
      when Invalid_Filename =>
         Send_Reply(Obj, The_550, Message_Invalid_Filename & Filename);
      when Event : Socket_Error=>
         -- the socket was closed by client
         Send_Reply(Obj, The_426, Message_Abort);
         Log(Obj.Task_Id, Exception_Information(Event));
         Close(File);
   end Command_Appe;

   -------------------------------------------------------------------------------------

   procedure Command_Dele (
         Obj      : Pi_Context_Type; 
         Filename : Unbounded_String ) is 

      Message_Deleted          : constant String := "File deleted.";  
      Message_Not_Allowed      : constant String := "Access not allowed.";  
      Message_Invalid_Filename : constant String := "Cannot delete file. Probably invalid filename.";  
   begin
      if Obj.Log_State = Logged then

         Delete_File(Obj.File_Tree, Filename);

         Send_Reply(Obj, The_250, Message_Deleted);
      else
         Send_Not_Logged_In(Obj);
      end if;

   exception
      when File_Trees.Access_Not_Allowed =>
         Send_Reply(Obj, The_550, Message_Not_Allowed);
      when Invalid_Filename =>
         Send_Reply(Obj, The_550, Message_Invalid_Filename);
   end Command_Dele;

   -------------------------------------------------------------------------------------

   procedure Command_Type (
         Obj    : in out Pi_Context_Type; 
         A_Type :        Unbounded_String ) is 

      Message_Set_To           : constant String := "Type set to: ";  
      Message_Invalid_Argument : constant String := "Argument expected for command TYPE.";  
      Message_Invalid_Type     : constant String := "This is not valid or supported type.";  
   begin
      if Obj.Log_State = Logged then

         if Length(A_Type) = 0 then
            -- we expect argument
            Send_Reply(Obj, The_501, Message_Invalid_Argument);
            return;
         end if;

         if Element(A_Type, 1) = 'i' or Element(A_Type, 1) = 'I' then
            -- the image type
            Send_Reply(Obj, The_200, Message_Set_To & "I - Image");
         elsif Element(A_Type, 1) = 'a' or Element(A_Type, 1) = 'A' then
            -- the ASCII type
            Send_Reply(Obj, The_200, Message_Set_To & "A - ASCII");
         else
            -- another types are unsupported
            Send_Reply(Obj, The_200, Message_Invalid_Type);
         end if;

      else
         Send_Not_Logged_In(Obj);
      end if;

   end Command_Type;

   -------------------------------------------------------------------------------------

   procedure Command_Mode (
         Obj    : in out Pi_Context_Type; 
         A_Mode :        Unbounded_String ) is 

      Message_Set_To           : constant String := "Mode set to: ";  
      Message_Invalid_Argument : constant String := "Argument expected for command MODE.";  
      Message_Invalid_Mode     : constant String := "This is not valid or supported mode.";  
   begin
      if Obj.Log_State = Logged then

         if Length(A_Mode) = 0 then
            -- we expect argument
            Send_Reply(Obj, The_501, Message_Invalid_Argument);
            return;
         end if;

         if Element(A_Mode, 1) = 's' or Element(A_Mode, 1) = 'S' then
            -- the stream mode
            Send_Reply(Obj, The_200, Message_Set_To & "S - Stream");
         else
            -- another types (blocks, compressed) are unsupported
            Send_Reply(Obj, The_200, Message_Invalid_Mode);
         end if;

      else
         Send_Not_Logged_In(Obj);
      end if;

   end Command_Mode;

   -------------------------------------------------------------------------------------

   procedure Command_Syst (
         Obj : Pi_Context_Type ) is 

      Message_Info : constant String := "UNIX Type: L8, Ada95; Probably running on WinNT but pretending unix :-)";  
      --Message_Info : constant String := "UNIX Type: L8";  -- testing string for fucking SmartFTP (client)
      --Message_Info : constant String := "Windows_NT";  
   begin
      Send_Reply(Obj, The_215, Message_Info);
   end Command_Syst;

   -------------------------------------------------------------------------------------

   procedure Command_Site (
         Obj : Pi_Context_Type ) is 
      Message_Superfluous : constant String := "Command not implemented, superfluous at this site.";  
   begin
      Send_Reply(Obj, The_202, Message_Superfluous);
   end Command_Site;

   -------------------------------------------------------------------------------------

   procedure Command_Noop (
         Obj : Pi_Context_Type ) is 

      Message_Noop_Ok : constant String := "Command OK. ""No operation"" succesfully done.";  
   begin
      Send_Reply(Obj, The_215, Message_Noop_Ok);
   end Command_Noop;

   -------------------------------------------------------------------------------------

   procedure Command_Help (
         Obj : Pi_Context_Type --; --Help_Subject : Unbounded_String 
         ) is 
      Message_Help : constant String := "Experimental FTP server written in Ada95; Copyright (c) Marek Paska 2002 - 2003";  
   begin
      Send_Reply(Obj, The_211, Message_Help);
   end;

   -------------------------------------------------------------------------------------

   procedure Command_Allo (
         Obj : Pi_Context_Type --; --Amount : Unbounded_String
         ) is 

      Message_Superfluous : constant String := "Command not implemented, superfluous at this site.";  
   begin
      Send_Reply(Obj, The_202, Message_Superfluous);
   end Command_Allo;

   -------------------------------------------------------------------------------------

   procedure Command_Acct (
         Obj : Pi_Context_Type ) is 

      Message_Superfluous : constant String := "Command not implemented, superfluous at this site.";  
   begin
      Send_Reply(Obj, The_202, Message_Superfluous);
   end Command_Acct;

   -------------------------------------------------------------------------------------

   procedure Command_Rest (
         Obj    : in out Pi_Context_Type; 
         Marker :        Unbounded_String ) is 

      Message_Restart_Marker_Set : constant String := "Restart marker set. Expecting STOR or RETR.";  
      Message_Syntax_Error       : constant String := "Parameter not recognized, not valid integer.";  
   begin
      if Obj.Log_State = Logged then
         Obj.Command_State := Restart_Marker_Set;
         Obj.Restart_Marker_Index := Positive_Count(Long_Long_Integer'
            Value(To_String(Marker)) +1);
         Send_Reply(Obj, The_350, Message_Restart_Marker_Set);
      else
         Send_Not_Logged_In(Obj);
      end if;
   exception
      when Constraint_Error =>
         Obj.Command_State := Uninteresting;
         Send_Reply(Obj, The_501, Message_Syntax_Error);
   end Command_Rest;

   -------------------------------------------------------------------------------------

   procedure Command_Mkd (
         Obj           : Pi_Context_Type; 
         New_Directory : Unbounded_String ) is 

      Message_Created                : constant String := " directory created.";  
      Message_Not_Allowed            : constant String := "Access not allowed.";  
      Message_Invalid_Directory_Name : constant String := "Invalid directory name or directory already exists: ";  
   begin
      if Obj.Log_State = Logged then
         Create_Directory(Obj.File_Tree, New_Directory);
         Send_Reply(Obj, The_257, """" & Get_Absolut_Name(Obj.File_Tree,
               New_Directory) & """" & Message_Created);
      else
         Send_Not_Logged_In(Obj);
      end if;

   exception
      when File_Trees.Access_Not_Allowed =>
         Send_Reply(Obj, The_550, Message_Not_Allowed);
      when Invalid_Filename =>
         Send_Reply(Obj, The_550, Message_Invalid_Directory_Name &
            Get_Absolut_Name(Obj.File_Tree, New_Directory));
   end Command_Mkd;

   -------------------------------------------------------------------------------------

   procedure Command_Rmd (
         Obj      : Pi_Context_Type; 
         Filename : Unbounded_String ) is 

      Message_Deleted          : constant String := "Directory deleted.";  
      Message_Not_Allowed      : constant String := "Access not allowed.";  
      Message_Invalid_Filename : constant String := "Cannot delete Directory. Probably invalid filename.";  
   begin
      if Obj.Log_State = Logged then

         Delete_Directory(Obj.File_Tree, Filename);

         Send_Reply(Obj, The_250, Message_Deleted);
      else
         Send_Not_Logged_In(Obj);
      end if;

   exception
      when File_Trees.Access_Not_Allowed =>
         Send_Reply(Obj, The_550, Message_Not_Allowed);
      when Invalid_Filename =>
         Send_Reply(Obj, The_550, Message_Invalid_Filename);

   end Command_Rmd;

   -------------------------------------------------------------------------------------

   procedure Command_Rnfr (
         Obj      : in out Pi_Context_Type; 
         Filename :        Unbounded_String ) is 

      Message_Waiting : constant String := "Command RNFR Ok. Expecting command RNTO.";  
   begin
      if Obj.Log_State = Logged then
         Obj.Command_State := Waiting_Rnto;
         Obj.Rename_From_Filename := Filename;
         Send_Reply(Obj, The_350, Message_Waiting);
      else
         Send_Not_Logged_In(Obj);
      end if;
   end Command_Rnfr;

   -------------------------------------------------------------------------------------
   procedure Command_Rnto (
         Obj      : in out Pi_Context_Type; 
         Filename :        Unbounded_String ) is 

      Message_Ok               : constant String := "File renamed.";  
      Message_Bad_Sequence     : constant String := "Bad sequence of commands. Expecting RNFR before RNTO.";  
      Message_Not_Allowed      : constant String := "Access not allowed.";  
      Message_Invalid_Filename : constant String := "Cannot rename file. Probably incorrect filename(s).";  
   begin
      if Obj.Log_State = Logged then
         if Obj.Command_State = Waiting_Rnto then
            Obj.Command_State := Uninteresting;

            Rename_File_Or_Directory(Obj.File_Tree, Obj.
               Rename_From_Filename, Filename);

            Send_Reply(Obj, The_250, Message_Ok);
         else
            Send_Reply(Obj, The_503, Message_Bad_Sequence);
         end if;
      else
         Send_Not_Logged_In(Obj);
      end if;

   exception
      when File_Trees.Access_Not_Allowed =>
         Send_Reply(Obj, The_553, Message_Not_Allowed);
      when Invalid_Filename =>
         Send_Reply(Obj, The_553, Message_Invalid_Filename);

   end Command_Rnto;

   -------------------------------------------------------------------------------------

   --
   -- Dispatch_Messages
   --

   procedure Dispatch_Message (
         Obj     : in out Pi_Context_Type;  
         Message :        Unbounded_String; 
         Quit    :    out Boolean           ) is 

      Command   : Pi_Commands.Command_Type;  
      Parameter : Unbounded_String;  

      Message_Unimplemented  : constant String := "Command not implemented.";  
      Message_Not_Understood : constant String := "Command not understood.";  
   begin
      Quit := False;
      Parse(Message, Command, Parameter);

      case Command is
         when The_User =>
            Command_User(Obj, Parameter);
         when The_Pass =>
            Command_Pass(Obj, Parameter);
         when The_Pwd | The_Xpwd =>
            Command_Pwd(Obj);
         when The_Cwd =>
            Command_Cwd(Obj, Parameter);
         when The_Cdup | The_Xcup =>
            Command_Cdup(Obj);
         when The_Port =>
            Command_Port(Obj, Parameter);
         when The_Type =>
            Command_Type(Obj, Parameter);
         when The_List =>
            Command_List(Obj, Parameter);
         when The_Nlst =>
            Command_List(Obj, Parameter, Names_Only => True);
         when The_Retr =>
            Command_Retr(Obj, Parameter);
         when The_Stor =>
            Command_Stor(Obj, Parameter);
         when The_Appe =>
            Command_Appe(Obj, Parameter);
         when The_Rest =>
            Command_Rest(Obj, Parameter);
         when The_Dele =>
            Command_Dele(Obj, Parameter);
         when The_Mkd | The_Xmkd =>
            Command_Mkd(Obj, Parameter);
         when The_Rmd | The_Xrmd =>
            Command_Rmd(Obj, Parameter);
         when The_Rnfr =>
            Command_Rnfr(Obj, Parameter);
         when The_Rnto =>
            Command_Rnto(Obj, Parameter);
         when The_Quit =>
            Command_Quit(Obj);
            Quit := True;
         when The_Syst =>
            Command_Syst(Obj);
         when The_Acct =>
            Command_Acct(Obj);
         when The_Allo =>
            Command_Allo(Obj--, Parameter
               );
         when The_Mode =>
            Command_Mode(Obj, Parameter);
         when The_Help =>
            Command_Help(Obj--, Parameter
               );
         when The_Noop =>
            Command_Noop(Obj);
         when The_Site =>
            Command_Site(Obj);
         when The_Pasv =>
            Command_Pasv(Obj);
         when others =>
            Send_Reply(Obj, The_502, Message_Unimplemented);
      end case;

      if Obj.Log_State = Logged then
         Obj.Statistics.Set_Activity(Obj.Task_Id, "None: " &
            Get_Current_Date & "    " & Get_Current_Time);
      end if;

   exception
      when Invalid_Command =>
         Send_Reply(Obj, The_500, Message_Not_Understood);
   end Dispatch_Message;


end Ftp_Protocol_Interpreter;

